# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
# None.

# Local imports
# None.


class ConfigurableParam:
    """
    A scooch parameter that is itself a scooch Configurable object.
    """

    def __init__(self, configurable_type, doc=""):
        """
        **Constructor**

        Args:
            configurable_type: Configurable, ConfigList or ConfigCollection - A scooch configurable type that
            specifies the class hierarchy for which this is a configuration for.

            doc: str - Docs describing what this param does in the context of its encapsulating Configurable
            class.
        """
        # TODO [matt.c.mccallum 11.04.21]: Ensure type is a configurable.
        self._type = configurable_type
        self._doc = doc

    @property
    def type(self):
        return self._type

    @property
    def doc(self):
        return f"<Configurable({str(self._type.__name__)})> - {self._doc}"

    def __set_name__(self, owner, name):
        """
        Assigns the name of the configurable. Any leading underscores are removed.

        Args:
            owner: Configurable - The scooch Configurable class that owns this ConfigurableParam.

            name: str - The name of the ConfigurableParam as an attribute in the owning Configurable.
        """
        self._name = name.lstrip('_')

    def __get__(self, instance, owner):
        """
        Returns the value of the parameter when the ConfigurableParam is accessed as an attribute
        of a Configurable class.

        Args:
            instance: Configurable - The instance of the class that this ConfigurableParam is an 
            attribute of.

            owner: Configurable - The class that this ConfigurableParam is a ConfigurableParam of.
        """
        # TODO [matt.c.mccallum 11.04.21]: Make sure owner is of type Configurable
        return instance._config_instances[self._name]
        


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
import click

# Local imports
# None.


def partial_option(*args, **kwargs):
    option = click.option(*args, **kwargs)

    def option_decorator(command=None):
        if command:
            # We were used as a decorator without arguments, and now we're being
            # called on the command function.
            return option(command)
        else:
            # We are being called with empty parens to construct a decorator.
            return option

    return option_decorator

#
# Options for several commands
#
config = partial_option(
        "--config", 
        "-c",
        help="Path to yaml scooch file for creation or processing"
    )

configurable = partial_option(
        "--configurable", 
        "-f",
        help="Complete import path to configurable class in python import syntax, e.g., package.module.classname", 
    )

pymodule = partial_option(
        "--pymodule", 
        "-m",
        help="Additional python packages / modules to import all from before running code.", 
        multiple=True,
)



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import logging
import textwrap

# Third party imports
# None.

# Local imports
# None.


def class_for_config(base_class, config):
    """
    Returns the feature class for a given config dictionary or Scooch Config object.
    This class search is performed under the assumption that each key in the dictionary
    or Config is a class in the pandafeet features module.

    Args:
        base_class: object - The base class for which you want to search for derived 
        classes of.

        config: dict - A dictionary with the keys specifying class names in the pandafeet
        features module.

    Return:
        class - The class that can be constructed with the provided config dictionary.
    """
    fclsses = base_class._all_subclasses() + [base_class]
    feature_class_names = [clsobj.__name__ for clsobj in fclsses]

    # Search for matching configuration
    config_fields = config.keys()
    matching_classes = [c_field for c_field in config_fields if c_field in feature_class_names]
    if not len(matching_classes):
        logging.getLogger().error(textwrap.dedent(f"""
                        Provided configuration does not match any, or matches multiple classes in the provided class hierarchy
                        Candidates were: {str(feature_class_names)}
                        Config requested: {str(list(config_fields))}
                        """))
        raise KeyError("""Scooch configuration does not match any, or matches multiple classes in the provided class hierarchy""")

    # Get all matching classes
    f_classes = [fclsses[feature_class_names.index(match_cls)] for match_cls in matching_classes]

    # Return one class if there is only one
    if len(f_classes) == 1:
        return f_classes[0]

    # Otherwise return the list of classes
    return f_classes

def class_instance_for_config(base_class, config):
    """
    Retrieve an instance of a class, constructed with the given configuration.

    Args:
        base_class: object - The base class for which you want to construct a derived class of.

        config: dict or Config - A configuration object specifying the class name (as the first key),
        and class config variables (as a dict in the first value), that you want to construct the class
        with. 
    
    Return:
        object - The constructed class instance that is a derived class of the base class and has been
        configured with the provided config.
    """
    # TODO [matt.c.mccallum 11.06.19]: Update other codebases to use this function where they can.
    return class_for_config(base_class, config)(config)
    


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import os
import json
import hashlib
import re
from datetime import datetime

# Third party module imports
import yaml

# Local imports
# None.


# NOTE [matt.c.mccallum 03.19.20]: The below constant is evaluated at module import time and
#      is required because datetime objects have trouble being pickled and sent to other machines
#      with other clock systems. The constant, evaluated at module import time, however, can
#      be pickled and distributed.
CURRENT_DATETIME = datetime.now().strftime("%Y%m%d_%H%M%S")


class Config(dict):
    """
    A simple class that encapsulates feature configuration files stored in yaml format.
    Each feature configuration file should contain all the details that make a feature unique for a certain type, e.g.,
    a CQT.
    """

    # TODO [matt.c.mccallum 07.03.19]: Allow loading a whole folder of configs
    # TODO [matt.c.mccallum 07.03.19]: Allow linking or references between configs to avoid duplication

    def __init__(self, config_file):
        """
        **Constructor.**

        Args:
            config_file: str, dict, file or Config - A file like object or filename describing a yaml file to load
            this class's configuration from. Alternatively, it can be a dictionary describing the 
        """
        if type(config_file) is str:
            with open(config_file) as f:
                self.update(yaml.safe_load(f))
        elif isinstance(config_file, dict):
            self.update(config_file)
        else:
            self.update(yaml.safe_load(config_file))

        # Store any variables
        if isinstance(config_file, Config):
            self._vars = config_file._vars
        else:
            self._vars = self.get('Constants', [])
            if len(self._vars):
                del self['Constants']

        self._VAR_FUNCS = {
            'inherit': Config._inherit,
            'datetime': Config._datetime
        }

        self._evaluate_vars(self)

    @classmethod
    def _inherit(cls, heirarchy, term, value):
        """
        Macro function to inherit a value from identical keys in it's parent dictionar(ies).

        Args:
            heirarchy: dict - The dictionary heirarchy containing the key and value pair
            that in turn contain a macro function.

            term: str - The dictionary key that the macro occurs at.

            value: str - The string containing the macro expression.

        Return:
            value: str - The provided value with the macro expression evaluated.
        """
        if 'PARENT' not in heirarchy.keys():
            raise KeyError('No inheritable attribute "{}" found in Scooch heirarchy'.format(term))
        return cls._inherit_recurse(heirarchy['PARENT'], term, value)

    @classmethod
    def _inherit_recurse(cls, heirarchy, term, value):
        """
        Recurses through all parent dicitonaries to find a matching key with whoms value
        to use to replace the provided value.

        See Config._inherit for details.
        """
        for child in heirarchy.keys():
            if child == term:
                return heirarchy[child]
        if 'PARENT' in heirarchy.keys():
            return cls._inherit_recurse(heirarchy['PARENT'], term, value)
        else:
            raise KeyError('No inheritable attribute "{}" found in Scooch heirarchy'.format(term))

    @classmethod
    def _datetime(cls, heirarchy, term, value):
        """
        Macro function to insert a datetime substring into the variable.

        Args:
            heirarchy: dict - The dictionary heirarchy containing the key and value pair
            that in turn contain a macro function.

            term: str - The dictionary key that the macro occurs at.

            value: str - The string containing the macro expression.

        Return:
            value: str - The provided value with the macro expression evaluated.
        """
        value = value.replace(r'${datetime}', CURRENT_DATETIME)
        return value

    def _evaluate_macro(self, heirarchy, term, value):
        """
        Evaluates all macros in the provided value.

        Args:
            heirarchy: dict - The dictionary heirarchy containing the key and value pair
            that in turn contain a macro function.

            term: str - The dictionary key that the macro occurs at.

            value: str - The string containing the macro expression.

        Return:
            value: str - The provided value with the macro expression evaluated.
        """
        while type(value) is str and len(re.findall(r'\$\{(.*?)\}', value)):
            func_name = re.findall(r'\$\{(.*?)\}', value)[0]  # NOTE [matt.c.mccallum 02.14.20]: Double evaluation of regex here, not the most efficient, but we're less concerned about efficiency in Scooch as it is not runtime.
            if func_name in self._vars:
                return self._vars[func_name]
            try:
                func = self._VAR_FUNCS[func_name]
            except KeyError:
                raise KeyError('No appropriate function "{}" defined in Scooch'.format(func_name))
            value = func(heirarchy, term, value)
        return value

    def _evaluate_vars(self, parent):
        """
        Performs a depth-first search through the dictionary heirarchy to
        find all macro functions and evaluate them, replacing the macros
        with the resulting values.
        Note that in a single string variable macros will be evaluated in 
        the order they occur in the string, so if one macro replaces the entire
        variable, or overwrites other macros, they will not be evaluated.

        Args:
            parent: dict - The current top level dictionary to recurse through
            and evaluate macros.
        """
        for child_key, child_var in parent.items():

            # Recurse for all children that are dictionaries
            if type(child_var) is dict and child_key is not 'PARENT':
                parent[child_key]['PARENT'] = parent
                self._evaluate_vars(parent[child_key])

            # If it is a list of values, treat it like each element in the list has the same
            # key and parent.
            elif type(child_var) is list and child_key is not 'PARENT':
                for idx, sub_var in enumerate(child_var):
                    if type(sub_var) is dict:
                        sub_var['PARENT'] = parent
                        self._evaluate_vars(sub_var)
                    elif type(sub_var) is str:
                        child_var[idx] = self._evaluate_macro(parent, child_key, sub_var)

            # If it is a string evaluate functions for all macros
            elif type(child_var) is str:
                parent[child_key] = self._evaluate_macro(parent, child_key, child_var)

        # Remove the reverse linkages to parents on the way out
        if 'PARENT' in parent.keys():
            del parent['PARENT']

    def save(self, config_file):
        """
        Saves a the configuration fields in yaml format to disk.

        Args:
            config_file: str or file - A file like object or filename describing a yaml file to save
            this class's configuration to.
        """
        if type(config_file) is str:
            with open(config_file, 'w') as f:
                f.write(yaml.safe_dump({k: v for k, v in self.items()}, default_flow_style=False))
        else:
            config_file.write(yaml.safe_dump({k: v for k, v in self.items()}, default_flow_style=False))
            
    @property
    def json(self):
        """
        Returns a string containing a json version of the dictionary that is unique for a unique configuration,
        and identical for identical configurations.

        Return:
            str - A json formatted string describing the dictionary contents
        """
        return json.dumps(self, sort_keys=True)

    @property
    def hashid(self):
        """
        Returns an identifier that will be the same for identical dictionary contents.
        Note there may be a hash collision here for two distinct contents, but it is very unlikely.

        Return:
            str - A string containing a hex representation of the hashed dictionary.
        """
        return hashlib.sha256(self.json.encode()).hexdigest()



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
# None.

# Local imports
from .configurable_meta import ConfigurableMeta
from .config_list import ConfigList
from .config_collection import ConfigCollection
from .helper_funcs import class_instance_for_config
from .config import Config


class ConfigurableFactory( object ):
    """
    A Factory for constructing scooch configurables for a given configuration.
    """

    def construct(self, cfg_type, cfg):
        """
        This will construct and return a Scooch configurable of a given cfg_type according to the parameterization
        specified in the cfg.

        Args:
            cfg_type: ConfigCollection, ConfigList or ConfigurableMeta class: This specifies the type to construct each
            scooch configurable as.

            cfg: scooch.Config: The config specifying the Configurable's parameterization.

        Return:
            scooch.Configurable (or list or dictionary thereof): The constructed scooch Configurable.
        """
        # TODO [matt.c.mccallum 01.05.21]: TYPE CHECK THAT THE CONFIG LISTS AND DICTS MATCH UP WITH THE `cfg_type` HERE.
        # TODO [matt.c.mccallum 11.04.21]: Raise a helpful error if we don't get a config list or config collection in the config dict, when we expect one.
        if isinstance(cfg_type, ConfigCollection):
            output_obj = {name: self.construct(cfg_type.subtype, cfg) for name, cfg in cfg.items()}
        elif isinstance(cfg_type, ConfigList):
            output_obj = [self.construct(cfg_type.subtype, each_cfg) for each_cfg in cfg]
        elif type(cfg_type) is ConfigurableMeta:
            # TODO [matt.c.mccallum 01.13.21]: Move the below method to a member function once we have updated all other codebases to use the ConfigurableFactory instead.
            output_obj = class_instance_for_config(cfg_type, Config(cfg))
        else:
            raise TypeError(f'{self.__class__.__name__} requested to construct class with Configurable parameter of type {str(type(cfg_type))}. Eligible types include ConfigList, ConfigCollection or a subclass of Configurable')
        return output_obj
        



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import copy

# Third party module imports
# None

# Local imports
from . import DEFAULT_NAMESPACE
from .configurable_meta import ConfigurableMeta
from .configurable_factory import ConfigurableFactory


class Configurable(object, metaclass=ConfigurableMeta):
    """
    A Base class for any object that has a given configuration, i.e., requires a certain set of parameters.

    The configuration may include a heirarchy. That is, in the list of required fields may be a dictionary
    with each key corresponding to a name of a sublist of feilds and each key being a list corresponding to 
    that sublist of field names.

    Note that anything below the top level of the heirarchy will not be assigned as an attribute.
    It is assumed that values further down the heirarchy are for configuration of other objects and it is
    up to the derived class to use these values.
    """

    __SCOOCH_NAME__ = None
    
    __PARAMS__ = {
        "config_namespace": "<str> - A namespace for the configuration, configs in distinct namespaces will have distinct identities."
    } # <= Parameters that can be specified in the class's configuration

    __CONFIGURABLES__ = {} # <= Parameters that are Scooch configurables and will be constructed according to the configuration dicts specified in the Config

    __PARAM_DEFAULTS__ = {
        "config_namespace": DEFAULT_NAMESPACE
    } # <= Parameters that are optional, and if not provided, will assume default values

    def __init__(self, cfg):
        """
        **Constructor.**

        Args:
            cfg: dict - An object providing the configuration parameters for this object.
        """
        # TODO [matt.c.mccallum 11.04.21]: Gracefully handle the error case that the top-level dictionary is not a single key value corresponding to a class...
        if not cfg[self.__class__.__name__]:
            self._cfg = {self.__class__.__name__: {}}
        else:
            # Save configuration dictionary
            self._cfg = cfg

        required_config = list(self.__PARAMS__.keys())

        # Populate defaults
        self._populate_defaults_recurse(self._cfg[self.__class__.__name__], self.__PARAM_DEFAULTS__)

        # Verify configuration
        # NOTE [matt.c.mccallum 12.16.20]: Just do this for this configurable, all sub-configurables will be verified upon their construction, respectively.
        self._verify_config(self._cfg[self.__class__.__name__], required_config)

        # Construct configurables
        self._config_factory = ConfigurableFactory()
        self._config_instances = {}
        for param_name, configurable in self.__CONFIGURABLES__.items():
            obj = self._config_factory.construct(configurable, self._cfg[self.__class__.__name__][param_name])
            self._config_instances[param_name] = obj

    @classmethod
    def populate_defaults(cls, cfg):
        """
        Populates the defaults for this Configurable in the provided cfg, without actually constructing it.

        Args:
            cfg: Config / Dict - A configuration for this configurable to be populated with defaults where missing
            config values exist.
        """
        cls._populate_defaults_recurse(cfg[cls.__name__], cls.__PARAM_DEFAULTS__)

    @classmethod
    def _populate_defaults_recurse(cls, cfg, defaults):
        """
        Iterates through all of the items in the defaults and transfers them over to the configuration
        if they are not already there.

        Args:
            cfg: dict() - An object to be used to configure a configurable class

            defaults: dict() - A set of defaults to configure in `cfg` if they do not already exist there.
        """
        # If no config was provided for this class, start with an empty dictionary.
        if cfg is None:
            cfg = dict()
        for field, value in defaults.items():
            if field not in cfg.keys():
                cfg[field] = value
            elif isinstance(value, dict):
                # Check the sub fields
                cls._populate_defaults_recurse(cfg[field], value)

    def _verify_config(self, cfg, template):
        """
        Checks that all the required fields in the object configuration are there.

        Args:
            cfg: dict() - An object providing the configuration parameters for this object.

            template: list() - A list of keys that describe the required fields to be configured for this object.
            This may includ dictionaries, allowing a heirarchy in the provided configuration.
        """
        # TODO [matt.c.mccallum 01.05.21]: ADD TYPE CHECKING HERE - CHECK ALL CONFIGURABLES ARE OF THE RIGHT TYPE,
        #      AND THOSE THAT ARE LISTS OF CONFIGURABLES SHOULD BE LISTS ALSO.
        for field in template:
            if isinstance(field, dict):
                for key in field.keys():
                    if key not in cfg.keys():
                        raise ValueError("Scooch config error: " + key + " value not found in " + self.__class__.__name__ + " object configuration")
                    self._verify_config(cfg[key], field[key])
            elif field not in cfg.keys():
                raise ValueError("Scooch config error: " + field + " value not found in " + self.__class__.__name__ + " object configuration")

    @property
    def cfg(self):
        """
        Getter function for a copy of this objects configuration.

        Return:
            dict{str:value} : A hierarchical dictionary of this class's configuration and all classes therein.
        """
        return copy.copy(self._cfg)

    @classmethod
    def _all_subclasses(cls):
        """
        cls.__subclasses__ only transcends one level of inheretance so this is a simple 
        recursion to find all subclasses.

        Returns:
            list(Configurable) - A list of all eligible configurables that derive from this class.
        """
        return list(set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in c._all_subclasses()]))



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import re
import textwrap

# Third party imports
from ruamel.yaml.comments import CommentedMap

# Local imports
from .config_list import ConfigList
from .config_collection import ConfigCollection


class ConfigFactory(object):
    """
    A class for making configurations for class hierarchies out of the box.
    """

    def __init__(self, interactive) -> None:
        """
        **Constructor**

        Args:
            interactive: bool - Whether to prompt user for input when constructing a configuration.
        """
        super().__init__()
        self._interactive = interactive

    def create_config(self, cls, level=0):
        """
        Generates a configuration for this configurable, with all default values filled out,
        comments in place, and sub-configurables selected by the user (or placeholders are
        inserted).

        Args:
            cls: Configurable - A Configurable class to construct a configuration for.

            level: int - How many layers deep in a configuration heirarchy we are. This can help with
            printing prompts in interactive mode.

        Returns: 
            Config - The constructed configurable with placeholders where no defaults exist.
        """
        config = CommentedMap()

        if self._interactive:
            print(textwrap.indent(f'===\nConfiguring Configurable: {cls.__name__}\n===\n', '  '*level))

        # Add defaults
        config.update(cls.__PARAM_DEFAULTS__)

        # Add configurables
        for param, cfgrble_type in cls.__CONFIGURABLES__.items():
            docs = cls.__PARAMS__[param]
            config[param] = self._populate_config(cfgrble_type, docs, level+1)
        
        # Add required parameters and comments
        for ky, val in cls.__PARAMS__.items():
            if ky not in config.keys():
                if re.match("\<([^>]+)\>", val):
                    typestr = re.match("\<([^>]+)\>", val).group(1)
                    config[ky] = f'<{typestr}>'
                else:
                    config[ky] = '<Unspecified Type>'
            config.yaml_add_eol_comment('<= '+val, ky, 65)

        return {cls.__name__: config}

    def _get_subconfig(self, c_type, docs, level):
        """
        Configure a sub configurable configuration, optionally with user prompts for types.

        Args:
            c_type: Configurable - The Configurable to have a configuration constructed for.

            docs: str - The docs to print out in user prompts to explain the config purpose to
            the user.

            level: int - How many layers deep in a configuration heirarchy we are. This can 
            help with printing prompts in interactive mode.
            
        Returns:
            dict - A dictionary containing the constructed configuration
        """
        formatted_docs = textwrap.indent('\n'.join(textwrap.wrap(docs, 80)), '  ')

        subclss = c_type._all_subclasses()
        if len(subclss) == 1:
            return self.create_config(subclss[0], level + 1)
        else:
            if self._interactive:
                inputting = True
                while inputting:
                    subcls_names = [f'{idx}: {subcls.__name__}' for idx, subcls in enumerate(subclss)]
                    print(textwrap.indent(f'Select Subclass for Component of Type "{c_type.__name__}":\n-\n{formatted_docs}\n-\n' + '\n'.join(subcls_names), '  '*level + '+ '))
                    try:
                        selection = int(input('  '*level + '+ '))
                        if selection < 0: raise IndexError  # Prevent negative indexing in UI.
                        print(' ')
                        inputting = False
                        return self.create_config(subclss[selection], level + 1)
                    except (ValueError, IndexError):
                        print(textwrap.indent(f'Invalid value, please enter an integer from 0 to {len(subclss)-1}', '  '*level))
                        print(' ')
            else:
                return {f'<{c_type.__name__}>': None}

    def _unpack_config_list(self, c_type, docs, level):
        """
        Configure a list of sub configurable configurations, optionally with user prompts 
        for types.

        Args:
            c_type: Configurable - The Configurable to have a configuration constructed for.

            docs: str - The docs to print out in user prompts to explain the config purpose to
            the user.

            level: int - How many layers deep in a configuration heirarchy we are. This can 
            help with printing prompts in interactive mode.

        Returns:
            dict - A dictionary containing the constructed configuration
        """
        formatted_docs = textwrap.indent('\n'.join(textwrap.wrap(docs, 80)), '  ')

        subtype = c_type.subtype
        if type(subtype) in (ConfigList, ConfigCollection):
            subtype_name = subtype.__class__.__name__
        else:
            subtype_name = subtype.__name__

        if self._interactive:
            inputting = True
            while inputting:
                
                print(textwrap.indent(f'Choose number of elements in Config List of type "{subtype_name}":\n-\n{formatted_docs}\n-', '  '*level + '+ '))
                try:
                    number_elem = int(input('  '*level + '+ '))
                    if number_elem < 0: raise ValueError
                    print(' ')
                    inputting = False
                except ValueError:
                    print(textwrap.indent(f'Invalid value, please enter a positive integer', '  '*level))
                    print(' ')
        cfg = []
        for _ in range(number_elem):
            if type(subtype) not in (ConfigList, ConfigCollection):
                docs = ''
            cfg += [self._populate_config(subtype, docs, level+1)]
        return cfg

    def _unpack_config_collection(self, c_type, docs, level):
        """
        Configure a collection of sub configurable configurations, optionally with user prompts 
        for types.

        Args:
            c_type: ConfigCollection - The Configurable Collection to have a configuration constructed for.

            docs: str - The docs to print out in user prompts to explain the config purpose to
            the user.

            level: int - How many layers deep in a configuration heirarchy we are. This can 
            help with printing prompts in interactive mode.

        Returns:
            dict - A dictionary containing the constructed configuration
        """
        formatted_docs = textwrap.indent('\n'.join(textwrap.wrap(docs, 80)), '  ')

        subtype = c_type.subtype
        if type(subtype) in (ConfigList, ConfigCollection):
            subtype_name = subtype.__class__.__name__
        else:
            subtype_name = subtype.__name__

        if self._interactive:
            inputting = True
            cfg = {}
            while inputting:
                print(textwrap.indent(f'Creating a collection of type "{subtype_name}":\n-\n{formatted_docs}\n-', '  '*level + '+ '))
                print(textwrap.indent(f'Choose a name for an element in collection of type "{subtype_name}", \nor press enter to finish generating this collection:', '  '*level + '+ '))
                name = input('  '*level + '+ ')
                print(' ')
                if len(name) == 0 or name.isspace():
                    inputting = False
                else:
                    if type(subtype) not in (ConfigList, ConfigCollection):
                        docs = ''
                    cfg[name] = self._populate_config(subtype, docs, level+1)
        return cfg

    def _populate_config(self, cfgrble_type, docs, lvl):
        """
        A hub function to distribute the configuration construction to the correct handler, based on the type
        of configuration required.

        Args:
            cfgrble_type: (Configurable, ConfigList, or ConfigCollection) - A type of configuration to be constructed.

            docs: str - The docs describing the configuration to be constructed.

            lvl: int - How deep in a configuration heirarchy we are - useful for pretty printing user prompts.
        """
        if type(cfgrble_type) is ConfigList:
            value = self._unpack_config_list(cfgrble_type, docs, lvl)
        elif type(cfgrble_type) is ConfigCollection:
            value = self._unpack_config_collection(cfgrble_type, docs, lvl)
        else:
            value = self._get_subconfig(cfgrble_type, docs, lvl)
        return value



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


DEFAULT_NAMESPACE = "root" # TODO [matt.c.mccallum 11.04.21]: <= Should be private

from enum import Enum
class ParamDefaults(Enum):
    NO_DEFAULT = 0

from .config import *
from .configurable import *
from .config_list import *
from .config_collection import *
from .config_factory import *
from .param import Param
from .configurable_param import ConfigurableParam

# TODO [matt.c.mccallum 01.05.21]: Remove the two below in favor of the config factory, once legacy codebases are updated.
from .configurize_helper import configurize
from .helper_funcs import class_for_config
from .helper_funcs import class_instance_for_config



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
# None.

# Local imports
# None.


class ConfigCollection(object):
    """
    A Scooch type that may be used to specify the type of sub-configurables a Scooch Configurable has.
    In this case it specifies an arbitrary length list of configurables.
    """

    def __init__(self, subtype):
        """
        **Constructor**

        Args:
            subtype: scooch.Configurable - The type of each Configurable in the dictionary.
        """
        self._subtype = subtype

    @property
    def subtype(self):
        """
        Retrieves the class specifying the type of Scooch Configurable expected in this dictionary.
        """
        return self._subtype
        


# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import importlib
import textwrap

# Third party imports
import click
import ruamel.yaml

# Local imports
from . import options
import json
from scooch import Configurable, config_factory
from scooch import ConfigFactory


@click.group()
def main():
    pass


@main.command()
@options.config
@options.configurable
@options.pymodule
def construct(config, configurable, pymodule):
    """
    Constructs a Config as a yaml file, for a given Configurable in a given module.
    """
    for module in pymodule:
        importlib.import_module(module)
    print('\n')
    try:
        configurable = next(x for x in Configurable._all_subclasses() if x.__name__ == configurable)
    except StopIteration:
        print(f"Error: No Configurable named {configurable} found (after importing: {pymodule})")
        return
    cfg = ConfigFactory(True).create_config(configurable)
    if config != None and len(config) > 0:
        with open(config, 'w') as f:
            ruamel.yaml.YAML().dump(cfg, f)
    else:
        print(json.dumps(cfg, indent=4, sort_keys=True))


@main.command()
@options.configurable
@options.pymodule
def options(configurable, pymodule):
    """
    Prints out all sub configurables for a given parent configurable.
    """
    for module in pymodule:
        importlib.import_module(module)
    try:
        configurable = next(x for x in Configurable._all_subclasses() if x.__name__ == configurable)
    except StopIteration:
        print(f"Error: No Configurable named {configurable} found (after importing: {pymodule})")
        return
    subclss = configurable._all_subclasses() + [configurable]
    # TODO [matt.c.mccallum 02.03.21]: If there's no subclasses just print out the base class
    names = [sub.__name__ for sub in subclss]
    docs = [sub.__doc__ for sub in subclss]
    result = ''
    for nm, dc in zip(names, docs):
        result += f'\n#\n# {nm}\n#\n'
        this_doc = dc.rstrip() + '\n\n'
        result += textwrap.indent(this_doc, '###', lambda x: True)
    print(result)


if __name__=='__main__':
    main()



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import inspect
import functools

# Third party imports
# None.

# Local imports
from .configurable import Configurable


def configurize(cls=None, base_class=None):
    """
    Takes a class and makes it scooch configurable. This will prepend "Conf" to the class name
    to distinguish it from the class definition. The returned / transformed class will be accessible
    via the name assigned in code, although it must be referred to as "Conf<original_class_name>" in
    scooch config files.

    Args:
        cls: class - A python class that will be made configurable via scooch.

        base_class: class - A python class that the newly minted scooch configurable class will inherit 
        from.

    Returns:
        class - The augmented Configurable class that may be configured via scooch.
    """

    def configurize_impl(cls, base_cls=None):

        # TODO [matt.c.mccallum 11.08.21]: Check the class is not already `Configurable`
        # TODO [matt.c.mccallum 11.08.21]: Check that base_cls is `Configurable`
        # TODO [matt.c.mccallum 11.08.21]: Inherit class documentation too

        if base_cls is None:
            base_cls = Configurable

        class DerivedConfigurable(cls, base_cls):
            """
            """

            __SCOOCH_NAME__ = 'Scooch' + cls.__name__

            # TODO [matt.c.mccallum 11.08.21]: Add type info here
            __PARAMS__ = {param: f'<> - Parameter derived by extending base class: {cls.__name__}' for param in inspect.signature(cls).parameters.keys()}

            __PARAM_DEFAULTS__ = {param: val.default for param, val in inspect.signature(cls).parameters.items() if val.default != val.empty}

            def __init__(self, cfg):
                cls.__init__(self, **(cfg[self.__class__.__name__]))
                Configurable.__init__(self, cfg)

        return DerivedConfigurable

    if base_class is None and cls is None:
        return None
    if base_class is None:
        return configurize_impl(cls)
    elif cls is None:
        return functools.partial(configurize_impl, base_cls=base_class)
    else:
        return configurize_impl(cls, base_class)





# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Local imports
# None.

# Third party imports
# None.

# Python standard library imports
import setuptools
import distutils.cmd


long_description = """
# Scooch

Scooch is a recursive acronym for **S**cooch **C**onfigures **O**bject 
**O**riented **C**lass **H**ierarchies, and that's exactly what this package 
does. It is a configuration package for python codebases that simplifies the 
problem of configuring parameters in python code by translating YAML 
configuration files into object oriented class hierarchies.

# Who needs Scooch?

Scooch is useful for people who need an accessible interface to enable 
tweakability in their code. ML practitioners are a good example. They 
typically write code that is intended to be continuously experimented with and 
adjusted in response to observations from running the code. As such, it is useful 
to abstract these tweakable parameters from the code into a config file, providing 
three major benefits:

 - The config file provides a centralized location for adjustable parameters of 
 interest in the code, improving iteration and workflow.
 - Loading, saving and adjusting the configuration of your code is separated 
 from the many other working variables and data structures that may exist in 
 code.
 - The configuration of any part of the code can be hashed, logged, and indexed, 
 to provide a record of the code configuration at any one time.

# Why use Scooch?

There are many other projects out there that endeavor to translate config files 
into parameters in python code, for example:

 - [Gin](https://github.com/google/gin-config)
 - [Sacred](https://sacred.readthedocs.io/en/stable/index.html)
 - [Hydra](https://hydra.cc/)

However, what makes Scooch different is that it not only translates config 
parameters into variables in your code, but into object oriented class 
hierarchies. This means configurations can benefit from object oriented concepts 
such as Inheretance, Encapsulation, Abstraction and Polymorphism.

For more information about how to use, and why to use Scooch. Please refer to 
the [documentation](http://www.mattcmccallum.com/scooch/docs).
"""


REQUIRED_PACKAGES = [
    'pyyaml==5.4.1',
    'sphinx',
    'sphinx_rtd_theme==0.5.1',
    'ruamel.yaml==0.16.12',
    'click==8.0.0a1'
]

class MakeReqsCommand(distutils.cmd.Command):
  """A custom command to export requirements to a requirements.txt file."""

  description = 'Export requirements to a requirements.txt file.'
  user_options = []

  def initialize_options(self):
    """Set default values for options."""
    pass

  def finalize_options(self):
    """Post-process options."""
    pass

  def run(self):
    """Run command."""
    with open('./requirements.txt', 'w') as f:
        for req in REQUIRED_PACKAGES:
            f.write(req)
            f.write('\n')


setuptools.setup(
    cmdclass={
        'make_reqs': MakeReqsCommand
    },
    name='scooch',
    version='0.1.0',
    description='A python module for configuring hierarchical class structures in yaml with defaults',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://www.mattcmccallum.com/scooch/docs",
    author="Matt C. McCallum",
    author_email="scooch@mattcmccallum.com",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={
        'Documentation': 'http://www.mattcmccallum.com/scooch/docs',
        'Bug Reports': 'https://github.com/PandoraMedia/scooch/issues',
        'Source': 'https://github.com/PandoraMedia/scooch',
    },
    license='Apache 2.0',
    keywords='scooch python configuration machine learning',

    install_requires=REQUIRED_PACKAGES,
    python_requires='>=3.6',
    packages=setuptools.find_packages(),
    include_package_data=True,

    # CLI
    entry_points = {
        'console_scripts': ['scooch=scooch.cli:main']
    }
)



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
import textwrap
import inspect

# Third party imports
# None.

# Local imports
from . import ParamDefaults
from .configurable_param import ConfigurableParam
from .param import Param


class ConfigurableMeta(type):
    """
    Metaclass for scooch configurables. Enables programmatic modification
    to the classes for things like programmatic class documentation.
    """

    def __new__(cls, name, bases, attrs):
        """
        **Preconstructor.**

        Currently does three things:
            - Collects all parameters and parameter defaults through the Configurable inheretence hierarchy
            - Translates scooch configuration dictionaries into scooch parameter types and vice versa
            - Updates the class doc string with information on its scooch parameters
        """
        # Get all base classes that are also ConfigurableMeta types
        meta_bases = [base for base in bases if type(base) is ConfigurableMeta]

        # Collect all params from base classes\
        cls._collect_param_from_bases(meta_bases, attrs, '__PARAMS__')
        cls._collect_param_from_bases(meta_bases, attrs, '__PARAM_DEFAULTS__')
        cls._collect_param_from_bases(meta_bases, attrs, '__CONFIGURABLES__')

        # Populate dictionaries with any scooch Param classes found on the class.
        for attr_name, value in attrs.items():

            if isinstance(value, Param):
                # Populate __PARAMS__
                attrs['__PARAMS__'][attr_name.lstrip('_')] = value.doc
                # Populate __PARAM_DEFAULTS__
                if value.default != ParamDefaults.NO_DEFAULT:
                    attrs['__PARAM_DEFAULTS__'][attr_name.lstrip('_')] = value.default

            if isinstance(value, ConfigurableParam):
                # Populate __PARAMS__
                attrs['__PARAMS__'][attr_name.lstrip('_')] = value.doc
                # Add to __CONFIGURABLES__
                attrs['__CONFIGURABLES__'][attr_name.lstrip('_')] = value.type

        # Create Param / ConfigurableParam attributes that don't already exist from the scooch Configurable dictionaries
        for attr_name in attrs['__PARAMS__']:
            if '_'+attr_name not in attrs and attr_name not in attrs:
                if attr_name in attrs['__CONFIGURABLES__']:
                    attrs['_'+attr_name] = ConfigurableParam(attrs['__CONFIGURABLES__'][attr_name], attrs['__PARAMS__'][attr_name])
                else:
                    attrs['_'+attr_name] = Param(None, attrs['__PARAM_DEFAULTS__'].get(attr_name, None), attrs['__PARAMS__'][attr_name])

        # Update the docs
        if not '__doc__' in attrs:
            attrs['__doc__'] = ""
        attrs['__doc__'] = cls._populate_docs(attrs['__doc__'], attrs['__PARAMS__'], attrs['__PARAM_DEFAULTS__'], attrs['__CONFIGURABLES__'])

        # Rename the class if programmatically defined
        if '__SCOOCH_NAME__' in attrs and attrs['__SCOOCH_NAME__'] is not None:
            name = attrs['__SCOOCH_NAME__']

        return super(ConfigurableMeta, cls).__new__(cls, name, bases, attrs)

    @staticmethod
    def _populate_docs(class_doc, params_dict, defaults_dict, configurables_dict):
        """
        Augments the class's doc string with information about the class's Scooch configuration.

        Args:
            class_doc: str - The class's doc string to be extended with Scooch config information.

            params_dict: dict(str, str) - The class's parameter dict, mapping parameter names to 
            parameter docs.

            defaults_dict: dict(str, object) - The class's defaults dict, mapping parameter names 
            to default parameter values

            configurables_dict: dict(str, Configurable) - The class's configurables dict, mapping
            Scooch configurable attributes to Scooch configurable types.

        Returns:
            class_doc: str - The class's doc string with Scooch configuration information appended.
        """

        if len(list(params_dict.keys())):
            class_doc += textwrap.indent(textwrap.dedent("""

            **Scooch Parameters**:
            """), '    ')

        for param, doc in params_dict.items():
            if param in list(configurables_dict.keys()):
                param_cls = configurables_dict[param]
                if inspect.isclass(param_cls):
                    default_info = f" (Configurable: {configurables_dict[param].__name__})"
                else:
                    # TODO [matt.c.mccallum 03.31.21]: Handle the case of ConfigList and ConfigCollection below...
                    default_info = ""
            elif param in list(defaults_dict.keys()):
                param_value = defaults_dict[param]
                if '\n' in str(param_value) or len(str(param_value)) > 40:
                    param_value = f" (Default is of type {type(param_value)})"
                else:
                    default_info = f" (Default: {defaults_dict[param]})"
            else:
                default_info = ""
            class_doc += textwrap.indent(textwrap.dedent(f"""
                                                **{param}**{default_info}:
                                                    {textwrap.fill(doc, 400)}
                                                """), '    ')

        return class_doc

    @staticmethod
    def _collect_param_from_bases(meta_bases, attrs, param_name):
        """
        Collects a parameter from the attributes of all the provided base classes.

        Args:
            meta_bases: list(Configurable) - A list of base classes from which to collect the parameters.
            This is to be provided in order of inheritance.

            attrs: dict - A dictionary of all attributes for this class that inherits the provided bases.

            param_name: str - The name of the parameter to collect across all provided base classes.
        """
        # Reverse list here to respect the python method resolution order (MRO) in any comprehension statements.
        meta_bases = meta_bases[::-1]
        # Collect the params.
        params = {k:v for base in meta_bases for k, v in base.__dict__[param_name].items()}
        if param_name in attrs.keys():
            params = {**params, **attrs[param_name]}
        attrs[param_name] = params



# coding=utf-8
# Copyright 2021 Pandora Media, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python standard library imports
# None.

# Third party imports
# None.

# Local imports
from . import ParamDefaults


class Param:
    """
    A Param that is not itself a scooch Configurable. For example, ints, floats, strings, or
    lists and dicts thereof.
    Params that are scooch Configurable class's should be of type ConfigurableParam.
    """

    def __init__(self, param_type, default=ParamDefaults.NO_DEFAULT, doc=""):
        """
        **Constructor**

        Args:
            param_type: object - The type of this parameter.

            default: object - The value of this Param, if it is not set in the scooch Config. This argument
            defaults to ParamDefaults.NO_DEFAULT, which will result in an error if the Param is missing from
            the scooch Config.

            doc: str - Docs describing what this param does in the context of its encapsulating Configurable
            class.
        """
        self._doc = doc
        self._type = param_type
        self._default = default

    @property
    def default(self):
        return self._default

    @property
    def doc(self):
        return f"<{str(self._type.__name__)}> - {self._doc}"

    def __set_name__(self, owner, name):
        """
        Assigns the name of the configurable. Any leading underscores are removed.

        Args:
            owner: Configurable - The scooch Configurable class that owns this Param.

            name: str - The name of this Param as an attribute in the owning Configurable.
        """
        self._name = name.lstrip('_')
    
    def __get__(self, instance, owner):
        """
        Returns the value of the parameter when the Param is accessed as an attribute
        of a Configurable class.

        Args:
            instance: Configurable - The instance of the class that this Param is an 
            attribute of.

            owner: Configurable - The class that this Param is a Param of.
        """
        # TODO [matt.c.mccallum 11.04.21]: Make sure owner is of type Configurable
        return instance._cfg[owner.__name__][self._name]
